@@ -20,6 +20,9 @@
 #include "shumate-map-layer.h"
 #include "shumate-memory-cache-private.h"
 #include "shumate-tile-private.h"
+#include "shumate-symbol-event.h"
+#include "shumate-profiling-private.h"
+#include "shumate-utils-private.h"
 
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
 #  include "vector/shumate-vector-symbol-container-private.h"
@@ -70,6 +73,14 @@ enum
 
 static GParamSpec *obj_properties[N_PROPERTIES] = { NULL, };
 
+enum
+{
+  SYMBOL_CLICKED,
+  LAST_SIGNAL
+};
+
+static guint signals[LAST_SIGNAL] = { 0, };
+
 /* This struct represents the location of a tile on the screen. It is the key
  * for the hash table tile_children which stores all visible tiles.
  *
@@ -77,60 +88,6 @@ static GParamSpec *obj_properties[N_PROPERTIES] = { NULL, };
  * here are *not* wrapped. For example, a ShumateTile at level 3 might have
  * coordinates of (7, 2) but have a TileGridPosition of (-1, 2). */
 
-typedef struct
-{
-  int x;
-  int y;
-  int zoom;
-} TileGridPosition;
-
-static void
-tile_grid_position_init (TileGridPosition *self,
-                         int x,
-                         int y,
-                         int zoom)
-{
-  self->x = x;
-  self->y = y;
-  self->zoom = zoom;
-}
-
-static TileGridPosition *
-tile_grid_position_new (int x,
-                        int y,
-                        int zoom)
-{
-  TileGridPosition *self = g_new0 (TileGridPosition, 1);
-  tile_grid_position_init (self, x, y, zoom);
-  return self;
-}
-
-static guint
-tile_grid_position_hash (gconstpointer pointer)
-{
-  const TileGridPosition *self = pointer;
-  return self->x ^ self->y ^ self->zoom;
-}
-
-static gboolean
-tile_grid_position_equal (gconstpointer a, gconstpointer b)
-{
-  const TileGridPosition *pos_a = a;
-  const TileGridPosition *pos_b = b;
-  return pos_a->x == pos_b->x && pos_a->y == pos_b->y && pos_a->zoom == pos_b->zoom;
-}
-
-static void
-tile_grid_position_free (gpointer pointer)
-{
-  TileGridPosition *self = pointer;
-  if (!self)
-    return;
-
-  g_free (self);
-}
-
-G_DEFINE_AUTOPTR_CLEANUP_FUNC (TileGridPosition, tile_grid_position_free);
 
 static int
 positive_mod (int i, int n)
@@ -142,7 +99,7 @@ typedef struct {
   ShumateMapLayer *self;
   ShumateTile *tile;
   char *source_id;
-  TileGridPosition pos;
+  ShumateGridPosition pos;
 } TileFilledData;
 
 static void
@@ -157,9 +114,9 @@ G_DEFINE_AUTOPTR_CLEANUP_FUNC (TileFilledData, tile_filled_data_free);
 
 
 static void
-add_symbols (ShumateMapLayer  *self,
-             ShumateTile      *tile,
-             TileGridPosition *pos)
+add_symbols (ShumateMapLayer     *self,
+             ShumateTile         *tile,
+             ShumateGridPosition *pos)
 {
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
   GPtrArray *symbols;
@@ -211,12 +168,15 @@ on_tile_filled (GObject      *source_object,
 }
 
 static void
-add_tile (ShumateMapLayer  *self,
-          ShumateTile      *tile,
-          TileGridPosition *pos)
+add_tile (ShumateMapLayer     *self,
+          ShumateTile         *tile,
+          ShumateGridPosition *pos)
 {
   const char *source_id = shumate_map_source_get_id (self->map_source);
 
+  self->profile_all_tiles_filled_begin = SHUMATE_PROFILE_CURRENT_TIME;
+  self->profile_all_tiles_done_begin = SHUMATE_PROFILE_CURRENT_TIME;
+
   if (shumate_memory_cache_try_fill_tile (self->memcache, tile, source_id))
     {
       add_symbols (self, tile, pos);
@@ -241,9 +201,9 @@ add_tile (ShumateMapLayer  *self,
 }
 
 static void
-remove_tile (ShumateMapLayer  *self,
-             ShumateTile      *tile,
-             TileGridPosition *pos)
+remove_tile (ShumateMapLayer     *self,
+             ShumateTile         *tile,
+             ShumateGridPosition *pos)
 {
   GCancellable *cancellable = g_hash_table_lookup (self->tile_fill, tile);
   if (cancellable)
@@ -259,6 +219,22 @@ remove_tile (ShumateMapLayer  *self,
   g_signal_handlers_disconnect_by_func (tile, on_tile_notify_state, self);
 }
 
+static double
+get_effective_zoom_level (ShumateMapLayer *self)
+{
+  double zoom_level = shumate_viewport_get_zoom_level (shumate_layer_get_viewport (SHUMATE_LAYER (self)));
+  ShumateMapSource *map_source = shumate_viewport_get_reference_map_source (shumate_layer_get_viewport (SHUMATE_LAYER (self)));
+
+  if (map_source)
+    {
+      double reference_tile_size = shumate_map_source_get_tile_size (map_source);
+      double our_tile_size = shumate_map_source_get_tile_size (self->map_source);
+      return log2 (reference_tile_size / our_tile_size) + zoom_level;
+    }
+  else
+    return zoom_level;
+}
+
 static gboolean
 defer_tick_callback (GtkWidget     *widget,
                      GdkFrameClock *frame_clock,
@@ -280,7 +256,7 @@ should_defer (ShumateMapLayer *self)
 
   ShumateViewport *viewport = shumate_layer_get_viewport (SHUMATE_LAYER (self));
   ShumateMapSource *map_source = self->map_source;
-  double zoom_level = shumate_viewport_get_zoom_level (viewport);
+  double zoom_level = get_effective_zoom_level (self);
   double tile_size = shumate_map_source_get_tile_size_at_zoom (map_source, zoom_level);
   double map_height = shumate_map_source_get_row_count (map_source, zoom_level) * tile_size;
   double map_width = shumate_map_source_get_column_count (map_source, zoom_level) * tile_size;
@@ -319,7 +295,7 @@ should_defer (ShumateMapLayer *self)
 
   self->defer_latitude_y = latitude_y / map_height;
   self->defer_longitude_x = longitude_x / map_width;
-  self->defer_zoom_level = shumate_viewport_get_zoom_level (viewport);
+  self->defer_zoom_level = zoom_level;
   self->defer_frame_time = frame_time;
 
   return self->deferring;
@@ -338,7 +314,7 @@ recompute_grid (ShumateMapLayer *self)
   int height = gtk_widget_get_height (GTK_WIDGET (self));
   ShumateViewport *viewport = shumate_layer_get_viewport (SHUMATE_LAYER (self));
   int tile_size = shumate_map_source_get_tile_size (self->map_source);
-  int zoom_level = shumate_viewport_get_zoom_level (viewport);
+  int zoom_level = (int)floor (get_effective_zoom_level (self));
   double latitude = shumate_location_get_latitude (SHUMATE_LOCATION (viewport));
   double longitude = shumate_location_get_longitude (SHUMATE_LOCATION (viewport));
   int latitude_y = shumate_map_source_get_y (self->map_source, zoom_level, latitude);
@@ -375,7 +351,7 @@ recompute_grid (ShumateMapLayer *self)
   g_hash_table_iter_init (&iter, self->tile_children);
   while (g_hash_table_iter_next (&iter, &key, &value))
     {
-      TileGridPosition *pos = key;
+      ShumateGridPosition *pos = key;
       ShumateTile *tile = value;
       float size = powf (2, zoom_level - pos->zoom);
       float x = pos->x * size;
@@ -397,7 +373,7 @@ recompute_grid (ShumateMapLayer *self)
     {
       for (int y = tile_initial_row; y < tile_initial_row + required_rows; y ++)
         {
-          g_autoptr(TileGridPosition) pos = tile_grid_position_new (x, y, zoom_level);
+          g_autoptr(ShumateGridPosition) pos = shumate_grid_position_new (x, y, zoom_level);
           ShumateTile *tile = g_hash_table_lookup (self->tile_children, pos);
 
           n_tiles ++;
@@ -405,6 +381,7 @@ recompute_grid (ShumateMapLayer *self)
           if (!tile && !defer)
             {
               tile = shumate_tile_new_full (positive_mod (x, source_columns), positive_mod (y, source_rows), tile_size, zoom_level);
+              shumate_tile_set_scale_factor (tile, gtk_widget_get_scale_factor (GTK_WIDGET (self)));
               add_tile (self, tile, g_steal_pointer (&pos));
             }
 
@@ -416,14 +393,27 @@ recompute_grid (ShumateMapLayer *self)
         }
     }
 
+  if (all_done && self->profile_all_tiles_done_begin > 0)
+    {
+      g_autofree char *desc = g_strdup_printf ("Visible tiles done (%d)", n_tiles);
+      SHUMATE_PROFILE_COLLECT (self->profile_all_tiles_done_begin, desc, NULL);
+      self->profile_all_tiles_done_begin = 0;
+    }
+  if (all_filled && self->profile_all_tiles_filled_begin > 0)
+    {
+      g_autofree char *desc = g_strdup_printf ("Visible tiles filled (%d)", n_tiles);
+      SHUMATE_PROFILE_COLLECT (self->profile_all_tiles_filled_begin, desc, NULL);
+      self->profile_all_tiles_filled_begin = 0;
+    }
+
   /* If all the tiles on the current zoom level are filled, delete tiles on all
    * other zoom levels */
-  if (all_filled)
+  if (all_done)
     {
       g_hash_table_iter_init (&iter, self->tile_children);
       while (g_hash_table_iter_next (&iter, &key, &value))
         {
-          TileGridPosition *pos = key;
+          ShumateGridPosition *pos = key;
           ShumateTile *tile = value;
 
           if (pos->zoom != zoom_level)
@@ -544,6 +534,16 @@ shumate_map_layer_dispose (GObject *object)
   G_OBJECT_CLASS (shumate_map_layer_parent_class)->dispose (object);
 }
 
+#ifdef SHUMATE_HAS_VECTOR_RENDERER
+static void
+on_symbol_clicked (ShumateMapLayer              *self,
+                   ShumateSymbolEvent           *event,
+                   ShumateVectorSymbolContainer *container)
+{
+  g_signal_emit (self, signals[SYMBOL_CLICKED], 0, event);
+}
+#endif
+
 static void
 shumate_map_layer_constructed (GObject *object)
 {
@@ -557,6 +557,7 @@ shumate_map_layer_constructed (GObject *object)
 
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
   self->symbols = shumate_vector_symbol_container_new (self->map_source, viewport);
+  g_signal_connect_object (self->symbols, "symbol-clicked", G_CALLBACK (on_symbol_clicked), self, G_CONNECT_SWAPPED);
   gtk_widget_set_parent (GTK_WIDGET (self->symbols), GTK_WIDGET (self));
 #endif
 }
@@ -596,28 +597,11 @@ shumate_map_layer_measure (GtkWidget      *widget,
                            int            *minimum_baseline,
                            int            *natural_baseline)
 {
-  ShumateMapLayer *self = SHUMATE_MAP_LAYER (widget);
-
   if (minimum)
     *minimum = 0;
 
   if (natural)
-    {
-      ShumateViewport *viewport;
-      double tile_size;
-      guint zoom_level;
-      guint count;
-
-      viewport = shumate_layer_get_viewport (SHUMATE_LAYER (self));
-      zoom_level = shumate_viewport_get_zoom_level (viewport);
-      tile_size = shumate_map_source_get_tile_size_at_zoom (self->map_source, zoom_level);
-      if (orientation == GTK_ORIENTATION_HORIZONTAL)
-        count = shumate_map_source_get_column_count (self->map_source, zoom_level);
-      else
-        count = shumate_map_source_get_row_count (self->map_source, zoom_level);
-
-      *natural = count * tile_size;
-    }
+    *natural = 0;
 }
 
 static double
@@ -633,7 +617,7 @@ shumate_map_layer_snapshot (GtkWidget *widget, GtkSnapshot *snapshot)
 {
   ShumateMapLayer *self = SHUMATE_MAP_LAYER (widget);
   ShumateViewport *viewport = shumate_layer_get_viewport (SHUMATE_LAYER (self));
-  double zoom_level = shumate_viewport_get_zoom_level (viewport);
+  double zoom_level = get_effective_zoom_level (self);
   int width = gtk_widget_get_width (GTK_WIDGET (self));
   int height = gtk_widget_get_height (GTK_WIDGET (self));
   double rotation = shumate_viewport_get_rotation (viewport);
@@ -688,7 +672,7 @@ shumate_map_layer_snapshot (GtkWidget *widget, GtkSnapshot *snapshot)
 
   while (g_hash_table_iter_next (&iter, &key, &value))
     {
-      TileGridPosition *pos = key;
+      ShumateGridPosition *pos = key;
       ShumateTile *tile = value;
       GdkPaintable *paintable = shumate_tile_get_paintable (tile);
       double size = tile_size * pow (2, zoom_level - pos->zoom);
@@ -752,12 +736,31 @@ shumate_map_layer_class_init (ShumateMapLayerClass *klass)
   g_object_class_install_properties (object_class,
                                      N_PROPERTIES,
                                      obj_properties);
+
+  /**
+   * ShumateMapLayer::symbol-clicked:
+   * @self: the [class@MapLayer] emitting the signal
+   * @event: a [class@SymbolEvent] with details about the clicked symbol.
+   *
+   * Emitted when a symbol in the map layer is clicked.
+   *
+   * Since: 1.1
+   */
+  signals[SYMBOL_CLICKED] =
+    g_signal_new ("symbol-clicked",
+                  G_OBJECT_CLASS_TYPE (object_class),
+                  G_SIGNAL_RUN_LAST,
+                  0, NULL, NULL,
+                  g_cclosure_marshal_VOID__OBJECT,
+                  G_TYPE_NONE,
+                  1,
+                  SHUMATE_TYPE_SYMBOL_EVENT);
 }
 
 static void
 shumate_map_layer_init (ShumateMapLayer *self)
 {
-  self->tile_children = g_hash_table_new_full (tile_grid_position_hash, tile_grid_position_equal, tile_grid_position_free, g_object_unref);
+  self->tile_children = g_hash_table_new_full (shumate_grid_position_hash, shumate_grid_position_equal, shumate_grid_position_free, g_object_unref);
   self->tile_fill = g_hash_table_new_full (g_direct_hash, g_direct_equal, g_object_unref, g_object_unref);
   self->memcache = shumate_memory_cache_new_full (100);
 }
